class ClientConf():
    def __init__(self, zk_host, config_groups, env, namespace):
        self.zk_host = zk_host
        self.config_groups = config_groups if isinstance(config_groups, (list, tuple)) else (config_groups,)
        self.env = env
        self.namespace = namespace
        self.recovery_mode = False



import re


def toPath(env,  namespace,  group_key,  meta_key):
    return f'/{namespace}/{env}/{group_key}/{meta_key}'


def group(env,  namespace,  group_key):
    return f'/{namespace}/{env}/{group_key}'


def toKey(meta_path: str):
    if not meta_path:
        return None
    return meta_path[meta_path.rindex('/')+1:]


def toNamespace(metaPath: str):
    if not metaPath:
        return None
    namespace = re.findall(r'^/([^/]+)/', metaPath)
    return namespace[0] if namespace else ''


def toGroup(metaPath: str):
    if not metaPath:
        return None
    metaPath = metaPath[:metaPath.rindex('/')]
    return metaPath[metaPath.rindex('/')+1:]



import enum
import logging

from pyfurion.config import ClientConf

logger = logging.getLogger(__name__)


class MetaEventType(enum.Enum):
    INIT = 1
    ADD = 2
    UPDATE = 3
    DELETE = 4


class MetaEvent():
    def __init__(self, env, event_type, namespace, config_group, meta_key, meta_value):
        self.env = env
        self.event_type = event_type
        self.namespace = namespace
        self.config_group = config_group
        self.meta_key = meta_key
        self.meta_value = meta_value


class MetaBucket():
    _storage = {}

    def get(self, meta_key: str):
        return self._storage.get(meta_key)

    def set(self, meta_key: str, meta_value: str):
        self._storage[meta_key] = meta_value

    def delete(self, meta_key: str):
        self._storage.pop(meta_key)


class MetaBucketFacade():
    _config_group_metabucket = {}

    def __init__(self, client_conf: ClientConf):
        for config_group in client_conf.config_groups:
            self.add_bucket(config_group, MetaBucket())

    def add_bucket(self, config_group, meta_bucket):
        self._config_group_metabucket[config_group] = meta_bucket

    def get_bucket(self, config_group) -> MetaBucket:
        return self._config_group_metabucket.get(config_group)

    def update(self, meta_event: MetaEvent):
        config_group = meta_event.config_group
        meta_bucket = self.get_bucket(config_group)
        if not config_group:
            return
        if meta_event.event_type in (MetaEventType.INIT, MetaEventType.UPDATE, MetaEventType.ADD):
            meta_bucket.set(meta_event.meta_key, meta_event.meta_value)
        elif meta_event.event_type == MetaEventType.DELETE:
            meta_bucket.delete(meta_event.meta_key, meta_event.meta_value)
        else:
            logger.warn(f'UNKNOW meta_event:{meta_event}')

    def get(self, config_group, meta_key):
        meta_bucket = self.get_bucket(config_group)
        if not config_group:
            return
        return meta_bucket.get(meta_key)



import logging


from pyfurion.config import ClientConf
from pyfurion.meta_bucket import MetaEvent, MetaBucketFacade
from pyfurion.snapshot import SnapshotProcess
from pyfurion.zookeeper import ZookeeperProcess

logger = logging.getLogger(__name__)

_namespace = 'furion'


class MetaListener():
    def meta_change(self, meta_evnet: MetaEvent):
        pass


class Client(MetaListener):
    def __init__(self, zk_host, config_groups, env, metachange_listeners):
        self.client_conf = ClientConf(zk_host, config_groups, env, _namespace)
        self._metachange_listeners = metachange_listeners
        self._meta_bucket_facade = MetaBucketFacade(self.client_conf)
        self._snapshop_process = SnapshotProcess(self.client_conf, self)
        self._zookeeper_process = ZookeeperProcess(self.client_conf, self)
        try:
            self._zookeeper_process.start()
        except Exception as e:
            logger.exception("furion metazookeeper Client启动异常")
            self.client_conf.recovery_mode = True
            self._snapshop_process.recovery()

    # def __getattr__(self, name: str) -> Any:
    #     return self.get(name)

    def get(self, meta_key, config_group=None, default=None):
        if not config_group:
            if len(self.client_conf.config_groups) == 1:
                config_group = self.client_conf.config_groups[0]
            else:
                raise RuntimeError('config_group can not be None if client inited by muti config_group')
        return self._meta_bucket_facade.get(config_group, meta_key) or default

    def meta_change(self, meta_event: MetaEvent):
        self._meta_bucket_facade.update(meta_event)
        self._snapshop_process.update(meta_event)
        self.dispatch_metaevent(meta_event)

    def dispatch_metaevent(self, meta_event: MetaEvent):
        if self._metachange_listeners:
            for meta_listener in self._metachange_listeners:
                meta_listener.meta_change(meta_event)



import json
import os


from pyfurion.config import ClientConf
from pyfurion.meta_bucket import MetaEvent, MetaEventType


class SnapshotProcess():
    def __init__(self, client_conf: ClientConf, meta_listener):
        self.root_dir = os.getenv('HOME', os.getcwd())
        self.meta_listener = meta_listener
        self.client_conf = client_conf
        for config_group in client_conf.config_groups:
            snapshot_path = self.get_snapshot_dir_path(config_group)
            if not os.path.exists(snapshot_path):
                os.makedirs(snapshot_path)

    def get_snapshot_dir_path(self, config_group):
        return f'{self.root_dir}/.{self.client_conf.namespace}/snapshot/{config_group}'

    def get_snapshot_meta_path(self, config_group, meta_key):
        return f'{self.get_snapshot_dir_path(config_group)}/{meta_key}.snapshot.json'

    def update(self, meta_event: MetaEvent):
        if self.client_conf.recovery_mode and meta_event.event_type == MetaEventType.INIT:
            return
        if meta_event.event_type != MetaEventType.DELETE:
            with open(self.get_snapshot_meta_path(meta_event.config_group, meta_event.meta_key), 'w') as f:
                f.write(json.dumps({
                    # 'id': meta_event.id,
                    'namespace': meta_event.namespace,
                    'configGroup': meta_event.config_group,
                    'metaKey': meta_event.meta_key,
                    'metaValue': meta_event.meta_value,
                    # 'version': meta_event.version,
                }))

    def recovery(self):
        for config_group in self.client_conf.config_groups:
            path = self.get_snapshot_dir_path(config_group)
            files = os.listdir(path)
            for file_name in files:
                index = file_name.rindex(".snapshot.json")
                if index < 0:
                    continue
                meta_key = file_name[:index]
                with open(f'{path}/{file_name}', 'r') as f:
                    meta_data = json.loads(f.read())
                    metaEvent = MetaEvent(self.client_conf.env, MetaEventType.INIT, self.client_conf.namespace,
                                          config_group, meta_key, meta_data['metaValue'])
                    self.dispatch_event(metaEvent)

    def dispatch_event(self, metaEvent: MetaEvent):
        self.meta_listener.meta_change(metaEvent)



import logging

from kazoo.client import KazooClient
from kazoo.protocol.states import (
    EventType)

from pyfurion import meta_path_util
from pyfurion.config import ClientConf
from pyfurion.meta_bucket import MetaEvent, MetaEventType

logger = logging.getLogger(__name__)


class ZookeeperProcess():
    def __init__(self, conf: ClientConf, meta_listener):
        self.conf = conf
        self.meta_listener = meta_listener
        self.sessionTimeoutMs = 15000
        self.connectionTimeoutMs = 15000
        self.baseSleepTimeMs = 1000
        self.maxRetries = 5
        self.zk = KazooClient(hosts=self.conf.zk_host)
        self.group_children = {}

    def start(self):
        self.zk.start()

        def data_watcher(event):
            logger.info(f'new event:{event}')
            _, namespace, env, config_group, meta_key = event.path.split('/')
            value = self.zk.get(event.path, watch=data_watcher)
            if value:
                value = value[0].decode()
            metaEvent = MetaEvent(self.conf.env, MetaEventType.INIT, namespace, config_group, meta_key, value)
            if event.type == EventType.CREATED:
                metaEvent.event_type = MetaEventType.ADD
            elif event.type == EventType.CHANGED:
                metaEvent.event_type = MetaEventType.UPDATE
            elif event.type == EventType.DELETED:
                metaEvent.event_type = MetaEventType.DELETE
            else:
                return
            self.dispatch_event(metaEvent)

        def init_child(config_group, name):
            path = meta_path_util.toPath(self.conf.env, self.conf.namespace, config_group, name)
            value = self.zk.get(path, watch=data_watcher)
            if value:
                value = value[0].decode()
            metaEvent = MetaEvent(self.conf.env, MetaEventType.INIT, self.conf.namespace,
                                  config_group, name, value)
            self.dispatch_event(metaEvent)

        for config_group in self.conf.config_groups:
            group = meta_path_util.group(self.conf.env, self.conf.namespace, config_group)
            self.zk.ensure_path(group)

            @self.zk.ChildrenWatch(group, send_event=True)
            def children_watcher(children, event):
                logger.info(f'children changed, new:{children},{event}')
                if not event:
                    return
                config_group = event.path[event.path.rindex('/') + 1:]
                # children = self.zk.get_children(e.path)
                current_children = self.group_children.get(config_group, [])
                added = set(children) - set(current_children)
                if added:
                    current_children.extend(added)
                    self.group_children[config_group] = current_children
                    for name in added:
                        init_child(config_group, name)

            children = self.zk.get_children(group)
            self.group_children[config_group] = children
            for name in children:
                init_child(config_group, name)

            logger.info(f"furion 开始监听path：{group}")
            logger.info(f"furion 初始化数据个数为：{len(children)}")

    def dispatch_event(self, metaEvent: MetaEvent):
        self.meta_listener.meta_change(metaEvent)



from setuptools import setup

'''
pip install wheel
python setup.py bdist_wheel
pip install twine
twine upload dist/*
'''

setup(
    # 以下为必需参数
    name='pyfurion',  # 模块名
    version='0.1.3',  # 当前版本
    description='A furion client for Python',  # 简短描述
    # py_modules=["pyfurion"],  # 单文件模块写法
    packages=['pyfurion'],

    # 以下均为可选参数
    long_description="",  # 长描述
    url='https://github.com',  # 主页链接
    author='zhangguanqun',  # 作者名
    author_email='zhanguanqunart@163.com',  # 作者邮箱
    classifiers=[
        "Development Status :: 5 - Production/Stable",

        'Intended Audience :: Developers',  # 模块适用人群
        'Topic :: Software Development :: Build Tools',  # 给模块加话题标签

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    keywords='furion client python',  # 模块的关键词，使用空格分割
    install_requires=['kazoo'],  # 依赖模块
    extras_require={  # 分组依赖模块，可使用pip install sampleproject[dev] 安装分组内的依赖
    },
    package_data={  # 模块所需的额外文件
    },
    data_files=[],  # 类似package_data, 但指定不在当前包目录下的文件
    project_urls={  # 项目相关的额外链接
        'Source': 'https://github.com',
    },
)
